const std = @import("std");
const Allocator = std.mem.Allocator;

const sdl = @import("sdl");
const keys = @import("./keys.zig");
const PhysicalKey = keys.PhysicalKey;
const Canvas = @import("./canvas.zig");

const Window = @This();

allocator: Allocator,

handle: *sdl.SDL_Window,
surface: ?*sdl.SDL_Surface = null,
key_status: []u8,
running: bool = true,

pub fn init(allocator: Allocator, width: usize, height: usize, key_status: []u8) !*Window {
    if (sdl.SDL_Init(sdl.SDL_INIT_AUDIO | sdl.SDL_INIT_VIDEO | sdl.SDL_INIT_EVENTS) < 0)
        return makeSdlError();

    const window = sdl.SDL_CreateWindow(
        "Gridstep",
        sdl.SDL_WINDOWPOS_UNDEFINED,
        sdl.SDL_WINDOWPOS_UNDEFINED,
        @intCast(width),
        @intCast(height),
        sdl.SDL_WINDOW_SHOWN | sdl.SDL_WINDOW_RESIZABLE,
    ) orelse
        return makeSdlError();

    const result = try allocator.create(Window);
    result.* = .{
        .allocator = allocator,
        .handle = window,
        .key_status = key_status,
    };
    return result;
}

pub fn deinit(self: *Window) void {
    sdl.SDL_DestroyWindow(self.handle);
    sdl.SDL_Quit();
    self.allocator.destroy(self);
}

pub fn pollEvents(self: *Window) void {
    {
        // Special handling for the `.caps` key.
        const i = @intFromEnum(PhysicalKey.caps);
        const on = (sdl.SDL_GetModState() & sdl.KMOD_CAPS) != 0;
        self.key_status[i] = if (on) i else 0;
    }

    var ev = std.mem.zeroes(sdl.SDL_Event);
    while (sdl.SDL_PollEvent(&ev) != 0) {
        switch (ev.type) {
            sdl.SDL_QUIT => {
                self.running = false;
            },
            sdl.SDL_WINDOWEVENT => {
                switch (ev.window.event) {
                    sdl.SDL_WINDOWEVENT_RESIZED => {
                        // Invalidate surface when window is resized.
                        self.surface = null;
                    },
                    else => {},
                }
            },
            sdl.SDL_KEYDOWN, sdl.SDL_KEYUP => {
                const key = ev.key.keysym;
                const down = (ev.type == sdl.SDL_KEYDOWN);
                if (keys.lookup[key.scancode]) |k| {
                    if (k == .caps) break;
                    const i = @intFromEnum(k);
                    self.key_status[i] = if (down) i else 0;
                }
            },
            else => {},
        }
    }
}

pub fn updateSurface(self: *Window, canvas: *Canvas) !void {
    // If surface has not yet been created or invalidated due
    // to window being resized, reaquire the window's surface.
    if (self.surface == null)
        self.surface = sdl.SDL_GetWindowSurface(self.handle) orelse
            return makeSdlError();

    const w: c_int = @intCast(canvas.width);
    const h: c_int = @intCast(canvas.height);
    const src_rect = sdl.SDL_Rect{ .x = 0, .y = 0, .w = w, .h = h };
    var dest_rect = sdl.SDL_Rect{ .x = 0, .y = 0, .w = self.surface.?.w, .h = self.surface.?.h };
    if (sdl.SDL_BlitScaled(canvas.surface, &src_rect, self.surface, &dest_rect) < 0)
        return makeSdlError();

    if (sdl.SDL_UpdateWindowSurface(self.handle) < 0)
        return makeSdlError();
}

pub fn makeSdlError() error{SdlError} {
    std.debug.print("{s}\n", .{sdl.SDL_GetError()});
    return error.SdlError;
}
